function [W,H] = isp_snmf2d(V, d, varargin)
% ISP_SNMF2D: SPARSE NON-NEGATIVE MATRIX FACTOR 2-D DECONVOLUTION
% 
% Usage:
%   [W,H] = sc2nmf(V,d,[options])
%
% Input:
%   V                 M x N data matrix
%   d                 Number of factors
%   options
%     .Tau            Shifts of H (default 0)
%     .Phi            Shifts of W (default 0)
%     .costfcn        Cost function to optimize
%                       'ls': Least squares (default)
%                       'kl': Kullback Leibler
%     .W              Initial W, array of size M x d x length(Tau)
%     .H              Initial H, array of size d x N x length(Phi) 
%     .lambda         Sparsity weight on H
%     .maxiter        Maximum number of iterations (default 100)
%     .conv_criteria  Function exits when cost/delta_cost exceeds this
%     .plotfcn        Function handle for plot function
%     .plotiter       Plot only every i'th iteration
%     .accel          Wild driver accelleration parameter (default 1.3)
%     .displaylevel   Level of display: [off | final | iter]
% 
% Output:
%   W                 M x d x length(Tau) array
%   H                 d x N x length(Phi) array
%
% Example I, Standard NMF:
%   d = 4;                                % Four components
%   [W,H] = isp_snmf2d(V,d);
%
% Example II, Sparse NMF2D:
%   d = 2;                                % Two components
%   opts.Tau = 0:7;                       % Eight time shifts
%   opts.Phi = 0:31;                      % Thirtytwo frequency shifts
%   opts.costfcn = 'kl';                  % Kullback Leibler cost function
%   opts.lambdaH = 10;                    % Sparsity
%   [W,H] = isp_snmf2d(V,d,opts);
% 
% Authors:
%   Mikkel N. Schmidt and Morten Mrup, 
%   Technical University of Denmark, 
%   Institute for Matematical Modelling
%
% References:
%   [1] M. Mrup and M. N. Schmidt. Sparse non-negative matrix factor 2-D 
%       deconvolution. Technical University of Denmark, 2006.
%   [2] M. N. Schmidt and M. Mrup. Nonnegative matrix factor 2-D 
%       deconvolution for blind single channel source separation. 
%       ICA, 2006.
%   [3] M. N. Schmidt and M. Mrup. Sparse non-negative matrix factor 2-d 
%       deconvolution for automatic transcription of polyphonic music. 
%       Submitted to EURASIP Journal on Applied Signal Processing, 2006.
%   [4] P. Smaragdis. Non-negative matrix factor deconvolution; 
%       extraction of multiple sound sourses from monophonic inputs. 
%       ICA 2004.
%   [5] J. Eggert and E. Korner. Sparse coding and nmf. In Neural Networks,
%       volume 4, 2004.
%   [6] J. Eggert, H. Wersing, and E. Korner. Transformation-invariant 
%       representation and nmf. In Neural Networks, volume 4, 2004.




% -------------------------------------------------------------------------
% Parse input arguments
if nargin>=3, opts = varargin{1}; else opts = struct; end
costfcn = mgetopt(opts, 'costfcn', 'ls', 'instrset', {'ls','kl'});
Tau = mgetopt(opts, 'Tau', 0);
Phi = mgetopt(opts, 'Phi', 0);
W = mgetopt(opts, 'W', rand(size(V,1),d,length(Tau)));
H = mgetopt(opts, 'H', rand(d,size(V,2),length(Phi)));
W = normalizeW(W);
nuH = mgetopt(opts, 'nuH', 1);
nuW = mgetopt(opts, 'nuW', 1);
lambda = mgetopt(opts, 'lambda', 0);
maxiter = mgetopt(opts, 'maxiter', 100);
conv_criteria = mgetopt(opts, 'conv_criteria', 1e-6);
accel = mgetopt(opts, 'accel', 1.3);
plotfcn = mgetopt(opts, 'plotfcn', []);
plotiter = mgetopt(opts, 'plotiter', 1);
displaylevel = mgetopt(opts, 'displaylevel', 'iter', 'instrset', ...
    {'off','iter','final'});



% -------------------------------------------------------------------------
% Initialization
sst = sum(sum((V-mean(mean(V(:)))).^2));
Rec = isp_nmf2d_rec(W,H,Tau,Phi,1:d); 
switch costfcn
    case 'ls'
        sse = norm(V-Rec,'fro')^2;
        cost_old = .5*sse + lambda*(sum(abs(H(:))));
    case 'kl'
        ckl = sum(sum(V.*log((V+eps)./(Rec+eps))-V+Rec));
        cost_old = ckl + lambda*(sum(abs(H(:))));
end
delta_cost = 1;
iter = 0;
keepgoing = 1;



% -------------------------------------------------------------------------
% Display information
dheader = sprintf('%12s | %12s | %12s | %12s | %12s | %12s','Iteration','Expl. var.','NuW','NuH','Cost func.','Delta costf.');
dline = sprintf('-------------+--------------+--------------+--------------+--------------+--------------');
if any(strcmp(displaylevel, {'final','iter'}))
    disp('Sparse Non-negative Matrix Factor 2-D Deconvolution');
    disp(['To stop algorithm press ctrl-C'])
    disp('');
end



% -------------------------------------------------------------------------
% Optimization loop
while keepgoing
    pause(0.001);
    
    if mod(iter,10)==0
        if any(strcmp(displaylevel, {'iter'}))
            disp(dline); disp(dheader); disp(dline);
        end
    end
    
    % Call plotfunction if specified
    if ~isempty(plotfcn) && mod(iter,plotiter)==0
        plotfcn(V,W,H,Tau,Phi,d); 
    end   
    
    % Update H and W
    switch costfcn
        case 'ls'
            [W,H,nuH,nuW,cost,sse,Rec] = ...
                ls_update(V,W,H,Tau,Phi,d,nuH,nuW,cost_old,accel,lambda,Rec);
        case 'kl'
            [W,H,nuH,nuW,cost,sse,Rec] = ...
                kl_update(V,W,H,Tau,Phi,d,nuH,nuW,cost_old,accel,lambda,Rec);
    end

    delta_cost = cost_old - cost;
    cost_old=cost;
    iter=iter+1;

    % Display information
    if any(strcmp(displaylevel, {'iter'}))
        disp(sprintf('%12.0f | %12.4f | %12.6f | %12.6f |  %11.2f | %11.2f', ...
            iter,(sst-sse)/sst,nuW/accel,nuH/accel,cost,delta_cost));
    end
    
    % Check if we should stop
    if delta_cost<cost*conv_criteria 
        % Small improvement with small step-size
        if nuH<=accel & nuW<=accel
            if any(strcmp(displaylevel, {'iter','final'}))
                disp('C2NMF has converged');
            end
            keepgoing = 0;
        % Small improvement - maybe because of too large step-size?
        else
            nuH = 1; nuW = 1;
        end
    end
    % Reached maximum number of iterations
    if iter>=maxiter
        if any(strcmp(displaylevel, {'iter','final'}))
            disp('Maximum number of iterations reached');
        end
        keepgoing=0; 
    end
end



% -------------------------------------------------------------------------
% Least squares update function
function [W,H,nuH,nuW,cost,sse,Rec] = ...
    ls_update(V,W,H,Tau,Phi,d,nuH,nuW,cost_old,accel,lambda,Rec)

H_old=H;
W_old=W;

% Update H
Hx = zeros(size(H));
Hy = zeros(size(H));
for tau = 1:length(Tau)   
    for phi = 1:length(Phi)
        Hx(:, 1:end-Tau(tau), phi) = Hx(:, 1:end-Tau(tau), phi) + ...
            W(1:end-Phi(phi), :, tau)'*V(1+Phi(phi):end,1+Tau(tau):end);
        Hy(:,1:end-Tau(tau),phi) = Hy(:,1:end-Tau(tau),phi) + ...
            W(1:end-Phi(phi), :, tau)'*Rec(1+Phi(phi):end,1+Tau(tau):end);          
    end
end
grad = Hx./(Hy+eps+lambda);
while 1
    H = H_old.*(grad.^nuH);
    Rec = isp_nmf2d_rec(W,H,Tau,Phi,1:d); 
    sse = norm(V-Rec,'fro')^2;
    cost = .5*sse+lambda*sum(H(:));
    if cost>cost_old, nuH = max(nuH/2,1);
    else nuH = nuH*accel; break; end
    pause(0);
end
cost_old = cost;

% Update W
Wx = zeros(size(W));
Wy = zeros(size(W));
for phi = 1:length(Phi)
    for tau = 1:length(Tau)   
        Wx(1:end-Phi(phi), : , tau) = Wx(1:end-Phi(phi), : , tau) + ...
            V(1+Phi(phi):end, 1+Tau(tau):end)*H(:, 1:end-Tau(tau), phi)';
        Wy(1:end-Phi(phi), : , tau) = Wy(1:end-Phi(phi), : , tau) + ...
            Rec(1+Phi(phi):end, 1+Tau(tau):end)*H(:, 1:end-Tau(tau), phi)';     
    end
end
tx = sum(sum(Wy.*W,1),3);
ty = sum(sum(Wx.*W,1),3);
Wx = Wx + repmat(tx,[size(W,1),1,size(W,3)]).*W;
Wy = Wy + repmat(ty,[size(W,1),1,size(W,3)]).*W;

grad = Wx./(Wy+eps);
while 1
    W = normalizeW(W_old.*(grad.^nuW));
    Rec = isp_nmf2d_rec(W,H,Tau,Phi,1:d); 
    sse = norm(V-Rec,'fro')^2;
    cost = .5*sse+lambda*sum(H(:));
    if cost>cost_old, nuW = max(nuW/2,1);
    else nuW = nuW*accel; break; end
    pause(0);
end
cost_old = cost;


% -------------------------------------------------------------------------
% Kullback Leibler update function
function [W,H,nuH,nuW,cost,sse,Rec] = ...
    kl_update(V,W,H,Tau,Phi,d,nuH,nuW,cost_old,accel,lambda,Rec)

H_old=H;
W_old=W;
VR = V./(Rec+eps);
O = ones(size(V));

% Update H
Hx = zeros(size(H));
Hy = zeros(size(H));
for tau = 1:length(Tau)   
    for phi = 1:length(Phi)
        Hx(:, 1:end-Tau(tau), phi) = Hx(:, 1:end-Tau(tau), phi) + ...
            W(1:end-Phi(phi), :, tau)'*VR(1+Phi(phi):end,1+Tau(tau):end);
        Hy(:,1:end-Tau(tau),phi) = Hy(:,1:end-Tau(tau),phi) + ...
            W(1:end-Phi(phi), :, tau)'*O(1+Phi(phi):end,1+Tau(tau):end);          
    end
end
grad = Hx./(Hy+eps+lambda);
while 1
    H = H_old.*(grad.^nuH);
    Rec = isp_nmf2d_rec(W,H,Tau,Phi,1:d); 
    ckl = sum(sum(V.*log((V+eps)./(Rec+eps))-V+Rec));
    cost = ckl + lambda*(sum(abs(H(:))));
    if cost>cost_old, nuH = max(nuH/2,1);
    else nuH = nuH*accel; break; end
    pause(0);
end
cost_old = cost;

VR = V./(Rec+eps);
% Update W
Wx = zeros(size(W));
Wy = zeros(size(W));
for phi = 1:length(Phi)
    for tau = 1:length(Tau)   
        Wx(1:end-Phi(phi), : , tau) = Wx(1:end-Phi(phi), : , tau) + ...
            VR(1+Phi(phi):end, 1+Tau(tau):end)*H(:, 1:end-Tau(tau), phi)';
        Wy(1:end-Phi(phi), : , tau) = Wy(1:end-Phi(phi), : , tau) + ...
            O(1+Phi(phi):end, 1+Tau(tau):end)*H(:, 1:end-Tau(tau), phi)';     
    end
end
tx = sum(sum(Wy.*W,1),3);
ty = sum(sum(Wx.*W,1),3);
Wx = Wx + repmat(tx,[size(W,1),1,size(W,3)]).*W;
Wy = Wy + repmat(ty,[size(W,1),1,size(W,3)]).*W;

grad = Wx./(Wy+eps);
while 1
    W = normalizeW(W_old.*(grad.^nuW));
    Rec = isp_nmf2d_rec(W,H,Tau,Phi,1:d); 
    ckl = sum(sum(V.*log((V+eps)./(Rec+eps))-V+Rec));
    cost = ckl + lambda*(sum(abs(H(:))));
    if cost>cost_old, nuW = max(nuW/2,1);
    else nuW = nuW*accel; break; end
    pause(0);
end
cost_old = cost;
sse = norm(V-Rec,'fro')^2;


% -------------------------------------------------------------------------
% Parser for optional arguments
function var = mgetopt(opts, varname, default, varargin)
if isfield(opts, varname)
    var = getfield(opts, varname); 
else
    var = default;
end
for narg = 1:2:nargin-4
    cmd = varargin{narg};
    arg = varargin{narg+1};
    switch cmd
        case 'instrset',
            if ~any(strcmp(arg, var))
                fprintf(['Wrong argument %s = ''%s'' - ', ...
                    'Using default : %s = ''%s''\n'], ...
                    varname, var, varname, default);
                var = default;
            end
        otherwise,
            error('Wrong option: %s.', cmd);
    end
end


% -------------------------------------------------------------------------
% Normalize W
function W = normalizeW(W)
Q(1,:,1) = sqrt(sum(sum(W.^2,1),3));
W = W./repmat(Q+eps,[size(W,1),1,size(W,3)]);
